.wy-side-nav-search {
    background-color: #fafafa;
    color: #404040;
}

.wy-side-nav-search > a {
    color: #2980B9;
}

.wy-side-nav-search > div.version {
    color: rgba(64, 64, 64, 0.3);
}

.red {
    color: red;
}

.green {
    color: green;
}